/*-----------------------------------------------------------------------

File  : scanner.c

Author: Stephan Schulz

Contents
 
  A simple scanner for PL

  Copyright 2002 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Sep 23 13:54:40 EDT 2002
    New

-----------------------------------------------------------------------*/

#include "scanner.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ScannerError()
//
//   Print an error message, including suspected position in the file.
//
// Global Variables: -
//
// Side Effects    : Output, terminates
//
/----------------------------------------------------------------------*/

void ScannerError(int line, int column, char *message)
{
   fprintf(stderr, "%s: %d (column %d) ", CurrentFileName(), line, column);
   Error(message);
}



/*-----------------------------------------------------------------------
//
// Function: SkipSpace()
//
//   Skip white space in the input.
//
// Global Variables: -
//
// Side Effects    : Reads input
//
/----------------------------------------------------------------------*/

void  SkipSpace(void)
{
   while(isspace(CurrentChar()))
   {
      NextChar();
   }
}

/*-----------------------------------------------------------------------
//
// Function: NextToken()
//
//   Return the next PL token. Not very elegant, but using only simple
//   C constructs (no passing of arrys...).
//
// Global Variables: -
//
// Side Effects    : Reads input
//
/----------------------------------------------------------------------*/

Token NextToken(void)
{
   Token tok;
   int i;

   SkipSpace();
   tok.line = CurrentLine();
   tok.column = CurrentColumn();
   tok.spelling[0] = '\0'; /* C convention: 0-terminated strings */
   tok.type = NoToken;

   switch(CurrentChar())
   {
   case 'X':
   case 'Y':
   case 'Z':
	 tok.type = Variable;
	 tok.spelling[0] = CurrentChar();
	 NextChar();
	 if(!isdigit(CurrentChar()))
	 {
	    ScannerError(tok.line, tok.column,"Incomplete Variable token");
	 }
	 if(CurrentChar() == '0')
	 {
	    ScannerError(tok.line, tok.column,"Non-Zero digit expected");
	 }
	 for(i=1; isdigit(CurrentChar()); i++)
	 {
	    if(i>=MAXLITLEN-1)
	    {
	       ScannerError(tok.line, tok.column,"Maximal token length reached");
	    }
	    tok.spelling[i] = CurrentChar();
	    NextChar();
	 }
	 tok.spelling[i] = '\0';
	 break;
   case 'l':
	 NextChar();
	 if(CurrentChar()!='o')
	 {
	    tok.type = Label;
	    tok.spelling[0] = 'l';
	    if(!isdigit(CurrentChar()))
	    {
	       ScannerError(tok.line, tok.column,"Incomplete label token");
	    }
	    if(CurrentChar() == '0')
	    {
	       ScannerError(tok.line, tok.column,"Non-Zero digit expected");
	    }
	    for(i=1; isdigit(CurrentChar()); i++)
	    {
	       if(i>=MAXLITLEN-1)
	       {
		  ScannerError(tok.line, tok.column,"Maximal token length reached");
	       }
	       tok.spelling[i] = CurrentChar();
	       NextChar();
	    }
	    tok.spelling[i] = '\0';
	    break;
	 }
	 /* It has to be "loop" */;
	 NextChar();
	 if(CurrentChar()!='o')
	 {
	    ScannerError(tok.line, tok.column,"Incomplete keyword");
	 }
	 NextChar();
	 if(CurrentChar()!='p')
	 {
	    ScannerError(tok.line, tok.column,"Incomplete keyword");
	 }
	 NextChar();
	 tok.type = ResWordLoop;
	 tok.spelling[0] = 'l';
	 tok.spelling[1] = 'o';
	 tok.spelling[2] = 'o';
	 tok.spelling[3] = 'p';
	 tok.spelling[4] = '\0';	    
	 break;
   case '0':
	 tok.type = NullDigit;
	 tok.spelling[0] = '0';
	 tok.spelling[1] = '\0';
	 NextChar();
	 break;
   case ';':
	 tok.type = Semicolon;
	 tok.spelling[0] = ';';
	 tok.spelling[1] = '\0';
	 NextChar();
	 break;
   case ':':
	 tok.type = Colon;
	 tok.spelling[0] = ':';
	 tok.spelling[1] = '\0';
	 NextChar();
	 break;
   case '+':
	 tok.type = IncOp;
	 tok.spelling[0] = '+';
	 tok.spelling[1] = '+';
	 tok.spelling[2] = '\0';
	 NextChar();
	 if(CurrentChar()!='+')
	 {
	    ScannerError(tok.line, tok.column,"Incomplete increment operator");
	 }
	 NextChar();
	 break;
   case '<':
	 tok.type = AssignOp;
	 tok.spelling[0] = '<';
	 tok.spelling[1] = '=';
	 tok.spelling[2] = '\0';
	 NextChar();
	 if(CurrentChar()!='=')
	 {
	    ScannerError(tok.line, tok.column,"Incomplete assignment operator");
	 }
	 NextChar();
	 break;   
   case 'e':
   case 'g':
	 for(i=0; isalpha(CurrentChar()); i++)
	 {
	    if(i>=MAXLITLEN-1)
	    {
	       ScannerError(tok.line, tok.column,"Maximal token Length reached");
	    }
	    tok.spelling[i] = CurrentChar();
	    NextChar();
	 }
	 tok.spelling[i] = '\0';
	 if(strcmp(tok.spelling, "endloop") == 0)
	 {
	    tok.type = ResWordEndloop;
	 }
	 else if(strcmp(tok.spelling, "goto") == 0)
	 {
	    tok.type = ResWordGoto;
	 }
	 else
	 {
	    ScannerError(tok.line, tok.column,"Invalid string");
	 }
	 
   case EOF:
	 /* Nothing to do */
	 break;
   default:
	 /* printf("'%c' <= \n", CurrentChar()); */
	 ScannerError(tok.line, tok.column,"Unexpected character");
	 break;
   }

   return tok;
}


/*-----------------------------------------------------------------------
//
// Function: TokenName()
//
//   Given a tokentype, return a matching string.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

char* TokenName(TokenType type)
{
   switch(type)
   {
   case NullDigit:
	 return   "NullDigit";
   case Semicolon:
	 return   "Semicolon";
   case Colon:
	 return   "Colon";
   case IncOp:
	 return   "IncOp";
   case AssignOp:
	 return   "AssignOp";
   case ResWordLoop:
	 return   "ResWordLoop";
   case ResWordEndloop:
	 return   "ResWordEndloop";
   case ResWordGoto:
	 return   "ResWordGoto";
   case Variable:
	 return   "Variable";
   case Label:
	 return   "Label";
   default: 
	 return   "NoToken";
   }
}


/*-----------------------------------------------------------------------
//
// Function: PrintToken()
//
//   Print a PL token
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void PrintToken(Token tok)
{
   printf("(%s, \"%s\", %d, %d)\n", TokenName(tok.type), 
	  tok.spelling, tok.line, tok.column);
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/








